# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from functools import wraps

from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import utils as sys_utils
from networking_cisco.apps.saf.db import dfa_db_models as dfa_dbm
from networking_cisco.apps.saf.server import dfa_openstack_helper as OsHelper
from networking_cisco.apps.saf.server.services.firewall.native import (
    fabric_setup_base as fabric)
from networking_cisco.apps.saf.server.services.firewall.native import (
    fw_constants)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    dev_mgr)

LOG = logging.getLogger(__name__)


class FwTenant(object):

    """Class to hold Tenant specific attributes.

    This class maintains a mapping of rule, policies, FW IDs with its
    associated tenant ID.
    It's assumed that a Rule, policy or FW signified by their unique ID can
    only be associated with one Tenant.
    """

    def __init__(self):
        """Initialize. """
        self.rule_tenant_dict = {}
        self.policy_tenant_dict = {}
        self.fw_tenant_dict = {}

    def store_rule_tenant(self, rule_id, tenant_id):
        """Stores the tenant ID corresponding to the rule. """
        self.rule_tenant_dict[rule_id] = tenant_id

    def get_rule_tenant(self, rule_id):
        """Retrieves the tenant ID corresponding to the rule. """
        return self.rule_tenant_dict[rule_id]

    def store_policy_tenant(self, policy_id, tenant_id):
        """Stores the tenant ID corresponding to the policy. """
        self.policy_tenant_dict[policy_id] = tenant_id

    def get_policy_tenant(self, policy_id):
        """Retrieves the tenant ID corresponding to the policy. """
        return self.policy_tenant_dict[policy_id]

    def store_fw_tenant(self, fw_id, tenant_id):
        """Stores the tenant ID corresponding to the firewall. """
        self.fw_tenant_dict[fw_id] = tenant_id

    def get_fw_tenant(self, fw_id):
        """Retrieves the tenant ID corresponding to the firewall. """
        return self.fw_tenant_dict[fw_id]

    def del_fw_tenant(self, fw_id):
        """Deletes the FW Tenant mapping. """
        del self.fw_tenant_dict[fw_id]

    def del_policy_tenant(self, pol_id):
        """Deletes the Tenant policy mapping. """
        del self.policy_tenant_dict[pol_id]

    def del_rule_tenant(self, rule_id):
        """Deletes the Tenant policy mapping. """
        del self.rule_tenant_dict[rule_id]


class FwMapAttr(object):

    """Firewall Attributes. This class is instantiated per tenant. """

    def __init__(self, tenant_id):
        """Initialize. """
        self.rules = {}
        self.policies = {}
        self.tenant_id = tenant_id
        self.rule_cnt = 0
        self.policy_cnt = 0
        self.active_pol_id = None
        self.fw_created = False
        self.fw_drvr_status = False
        self.fw_id = None
        self.fw_type = None
        self.tenant_name = None
        self.fw_name = None
        self.mutex_lock = sys_utils.lock()

    def store_policy(self, pol_id, policy):
        """Store the policy.

        Policy is maintained as a dictionary of pol ID.
        """
        if pol_id not in self.policies:
            self.policies[pol_id] = policy
            self.policy_cnt += 1

    def store_rule(self, rule_id, rule):
        """Store the rules.

        Policy is maintained as a dictionary of Rule ID.
        """
        if rule_id not in self.rules:
            self.rules[rule_id] = rule
            self.rule_cnt += 1

    def delete_rule(self, rule_id):
        """Delete the specific Rule from dictionary indexed by rule id. """
        if rule_id not in self.rules:
            LOG.error("No Rule id present for deleting %s", rule_id)
            return
        del self.rules[rule_id]
        self.rule_cnt -= 1
        # No need to navigate through self.policies to delete rules since
        # if a rule is a part of a policy, Openstack would not allow to delete
        # that rule

    def is_rule_present(self, rule_id):
        """Returns if rule specified by rule id is present in dictionary. """
        if rule_id not in self.rules:
            return False
        else:
            return True

    def rule_update(self, rule_id, rule):
        """Update the rule. """
        if rule_id not in self.rules:
            LOG.error("Rule ID not present %s", rule_id)
            return
        self.rules[rule_id].update(rule)

    def is_policy_present(self, pol_id):
        """Returns if policy specified by ID is present in the dictionary. """
        return pol_id in self.policies

    def is_fw_present(self, fw_id):
        """Returns if firewall index by ID is present in dictionary. """
        if self.fw_id is None or self.fw_id != fw_id:
            return False
        else:
            return True

    def create_fw(self, proj_name, pol_id, fw_id, fw_name, fw_type, rtr_id):
        """Fills up the local attributes when FW is created. """
        self.tenant_name = proj_name
        self.fw_id = fw_id
        self.fw_name = fw_name
        self.fw_created = True
        self.active_pol_id = pol_id
        self.fw_type = fw_type
        self.router_id = rtr_id

    def delete_fw(self, fw_id):
        """Deletes the FW local attributes. """
        self.fw_id = None
        self.fw_name = None
        self.fw_created = False
        self.active_pol_id = None

    def delete_policy(self, pol_id):
        """Deletes the policy from the local dictionary. """
        if pol_id not in self.policies:
            LOG.error("Invalid policy %s", pol_id)
            return
        del self.policies[pol_id]
        self.policy_cnt -= 1

    def is_fw_complete(self):
        """This API returns the completion status of FW.

        This returns True if a FW is created with a active policy that has
        more than one rule associated with it and if a driver init is done
        successfully.
        """
        LOG.info("In fw_complete needed %(fw_created)s "
                 "%(active_policy_id)s %(is_fw_drvr_created)s "
                 "%(pol_present)s %(fw_type)s",
                 {'fw_created': self.fw_created,
                  'active_policy_id': self.active_pol_id,
                  'is_fw_drvr_created': self.is_fw_drvr_created(),
                  'pol_present': self.active_pol_id in self.policies,
                  'fw_type': self.fw_type})
        if self.active_pol_id is not None:
            LOG.info("In Drvr create needed %(len_policy)s %(one_rule)s",
                     {'len_policy':
                      len(self.policies[self.active_pol_id]['rule_dict']),
                      'one_rule':
                      self.one_rule_present(self.active_pol_id)})
        return self.fw_created and self.active_pol_id and (
            self.is_fw_drvr_created()) and self.fw_type and (
            self.active_pol_id in self.policies) and (
            len(self.policies[self.active_pol_id]['rule_dict'])) > 0 and (
            self.one_rule_present(self.active_pol_id))

    def is_fw_drvr_create_needed(self):
        """This API returns True if a driver init needs to be performed.

        This returns True if a FW is created with a active policy that has
        more than one rule associated with it and if a driver init is NOT
        done.
        """
        LOG.info("In Drvr create needed %(fw_created)s "
                 "%(active_policy_id)s"
                 " %(is_fw_drvr_created)s %(pol_present)s %(fw_type)s",
                 {'fw_created': self.fw_created,
                  'active_policy_id': self.active_pol_id,
                  'is_fw_drvr_created': self.is_fw_drvr_created(),
                  'pol_present': self.active_pol_id in self.policies,
                  'fw_type': self.fw_type})
        if self.active_pol_id is not None and (
           self.active_pol_id in self.policies):
            LOG.info("In Drvr create needed %(len_policy)s %(one_rule)s",
                     {'len_policy':
                      len(self.policies[self.active_pol_id]['rule_dict']),
                      'one_rule':
                      self.one_rule_present(self.active_pol_id)})
        return self.fw_created and self.active_pol_id and (
            not self.is_fw_drvr_created()) and self.fw_type and (
            self.active_pol_id in self.policies) and (
            len(self.policies[self.active_pol_id]['rule_dict'])) > 0 and (
            self.one_rule_present(self.active_pol_id))

    def one_rule_present(self, pol_id):
        """Returns if atleast one rule is present in the policy. """
        pol_dict = self.policies[pol_id]
        for rule in pol_dict['rule_dict']:
            if self.is_rule_present(rule):
                return True
        return False

    def fw_drvr_created(self, status):
        """This stores the status of the driver init.

        This API assumes only one FW driver.
        """
        self.fw_drvr_status = status

    def is_fw_drvr_created(self):
        """This returns the status of the driver creation.

        This API assumes only one FW driver.
        """
        return self.fw_drvr_status

    def get_fw_dict(self):
        """This API creates a FW dictionary from the local attributes. """
        fw_dict = {}
        if self.fw_id is None:
            return fw_dict
        fw_dict = {'rules': {}, 'tenant_name': self.tenant_name,
                   'tenant_id': self.tenant_id, 'fw_id': self.fw_id,
                   'fw_name': self.fw_name,
                   'firewall_policy_id': self.active_pol_id,
                   'fw_type': self.fw_type, 'router_id': self.router_id}
        # When Firewall and Policy are both deleted and the SM is doing a
        # retry (maybe DCNM Out partition could not be deleted) during
        # which without this check, it throws an exception since
        # self.policies is empty. This is also an issue during restart.
        if self.active_pol_id not in self.policies:
            return fw_dict
        pol_dict = self.policies[self.active_pol_id]
        for rule in pol_dict['rule_dict']:
            fw_dict['rules'][rule] = self.rules[rule]
        return fw_dict

    def update_fw_params(self, rtr_id=-1, fw_type=-1):
        """Updates the FW parameters. """
        if rtr_id != -1:
            self.router_id = rtr_id
        if fw_type != -1:
            self.fw_type = fw_type


class FwMgr(dev_mgr.DeviceMgr):

    """Firewall Native Manager"""

    def __init__(self, cfg):
        """Initialization routine.

        It populates the local cache after reading the DB. It initializes the
        Fabric class and DeviceMgr class.
        """
        LOG.debug("Initializing Native FW Manager %s", cfg.firewall.device)
        self.fw_init = False
        if cfg.firewall.device is None or cfg.firewall.device is '':
            return
        super(FwMgr, self).__init__(cfg)
        self.events.update({
            'firewall_rule.create.end': self.fw_rule_create,
            'firewall_rule.delete.end': self.fw_rule_delete,
            'firewall_rule.update.end': self.fw_rule_update,
            'firewall_policy.create.end': self.fw_policy_create,
            'firewall_policy.delete.end': self.fw_policy_delete,
            'firewall.create.end': self.fw_create,
            'firewall.update.end': self.fw_update,
            'firewall.delete.end': self.fw_delete})
        self.fwid_attr = {}
        self.pid_dict = {}
        self.rules_id = {}
        self.fw_drvr_created = False
        self.fabric = fabric.FabricBase()
        self.tenant_db = FwTenant()
        self.os_helper = OsHelper.DfaNeutronHelper()
        fw_dict = self.populate_local_cache()
        self.populate_local_sch_cache(fw_dict)
        self.dcnm_obj = None
        self.fw_init = True

    def populate_cfg_dcnm(self, cfg, dcnm_obj):
        """This routine stores the DCNM object. """
        if not self.fw_init:
            return
        self.dcnm_obj = dcnm_obj
        self.fabric.store_dcnm(dcnm_obj)
        self.populate_dcnm_obj(dcnm_obj)

    def populate_event_queue(self, cfg, que_obj):
        """This routine is for storing the Event Queue obj. """
        if not self.fw_init:
            return
        self.que_obj = que_obj
        self.populate_event_que(que_obj)

    def network_sub_create_notif(self, tenant_id, tenant_name, cidr):
        """Network create notification. """
        if not self.fw_init:
            return
        self.network_create_notif(tenant_id, tenant_name, cidr)

    def network_del_notif(self, tenant_id, tenant_name, net_id):
        """Network delete notification. """
        if not self.fw_init:
            return
        self.network_delete_notif(tenant_id, tenant_name, net_id)

    def project_create_notif(self, tenant_id, tenant_name):
        """Tenant Create notification. """
        if not self.fw_init:
            return
        self.os_helper.create_router('_'.join([fw_constants.TENANT_EDGE_RTR,
                                               tenant_name]),
                                     tenant_id, [])

    def project_delete_notif(self, tenant_id, tenant_name):
        """Tenant Delete notification. """
        if not self.fw_init:
            return
        rtr_name = '_'.join([fw_constants.TENANT_EDGE_RTR, tenant_name])
        self.os_helper.delete_router_by_name(rtr_name, tenant_id)

    def _create_fw_fab_dev_te(self, tenant_id, drvr_name, fw_dict):
        """Prepares the Fabric and configures the device.

        This routine calls the fabric class to prepare the fabric when
        a firewall is created. It also calls the device manager to
        configure the device. It updates the database with the final
        result.
        """
        is_fw_virt = self.is_device_virtual()
        ret = self.fabric.prepare_fabric_fw(tenant_id, fw_dict, is_fw_virt,
                                            fw_constants.RESULT_FW_CREATE_INIT)

        if not ret:
            LOG.error("Prepare Fabric failed")
            return
        else:
            self.update_fw_db_final_result(fw_dict.get('fw_id'), (
                fw_constants.RESULT_FW_CREATE_DONE))
        ret = self.create_fw_device(tenant_id, fw_dict.get('fw_id'),
                                    fw_dict)
        if ret:
            self.fwid_attr[tenant_id].fw_drvr_created(True)
            self.update_fw_db_dev_status(fw_dict.get('fw_id'), 'SUCCESS')
            LOG.info("FW device create returned success for tenant %s",
                     tenant_id)
        else:
            LOG.error("FW device create returned failure for tenant %s",
                      tenant_id)

    def _create_fw_fab_dev(self, tenant_id, drvr_name, fw_dict):
        """This routine calls the Tenant Edge routine if FW Type is TE. """
        if fw_dict.get('fw_type') == fw_constants.FW_TENANT_EDGE:
            self._create_fw_fab_dev_te(tenant_id, drvr_name, fw_dict)

    def _check_create_fw(self, tenant_id, drvr_name):
        """Creates the Firewall, if all conditions are met.

        This function first checks if all the configuration are done
        for a FW to be launched. After that it creates the FW entry in the
        DB. After that, it calls the routine to prepare the fabric and
        configure the device.
        """
        if self.fwid_attr[tenant_id].is_fw_drvr_create_needed():
            fw_dict = self.fwid_attr[tenant_id].get_fw_dict()
            try:
                with self.fwid_attr[tenant_id].mutex_lock:
                    ret = self.add_fw_db(fw_dict.get('fw_id'), fw_dict,
                                         fw_constants.RESULT_FW_CREATE_INIT)
                    if not ret:
                        LOG.error("Adding FW DB failed for tenant %s",
                                  tenant_id)
                        return
                    self._create_fw_fab_dev(tenant_id, drvr_name, fw_dict)
            except Exception as exc:
                LOG.error("Exception raised in create fw %s", str(exc))

    def _delete_fw_fab_dev(self, tenant_id, drvr_name, fw_dict):
        """Deletes the Firewall.

        This routine calls the fabric class to delete the fabric when
        a firewall is deleted. It also calls the device manager to
        unconfigure the device. It updates the database with the final
        result.
        """
        is_fw_virt = self.is_device_virtual()
        if self.fwid_attr[tenant_id].is_fw_drvr_created():
            ret = self.delete_fw_device(tenant_id, fw_dict.get('fw_id'),
                                        fw_dict)
            if not ret:
                LOG.error("Error in delete_fabric_fw device for tenant "
                          "%s", tenant_id)
                return False
            else:
                self.fwid_attr[tenant_id].fw_drvr_created(False)
                self.update_fw_db_dev_status(fw_dict.get('fw_id'), '')
        ret = self.fabric.delete_fabric_fw(tenant_id, fw_dict, is_fw_virt,
                                           fw_constants.RESULT_FW_DELETE_INIT)
        if not ret:
            LOG.error("Error in delete_fabric_fw for tenant %s",
                      tenant_id)
            return False
        self.update_fw_db_final_result(fw_dict.get('fw_id'), (
            fw_constants.RESULT_FW_DELETE_DONE))
        self.delete_fw(fw_dict.get('fw_id'))
        return True

    def _check_delete_fw(self, tenant_id, drvr_name):
        """Deletes the Firewall, if all conditioms are met.

        This function after modifying the DB with delete operation status,
        calls the routine to remove the fabric cfg from DB and unconfigure
        the device.
        """
        fw_dict = self.fwid_attr[tenant_id].get_fw_dict()
        ret = False
        try:
            with self.fwid_attr[tenant_id].mutex_lock:
                self.update_fw_db_final_result(fw_dict.get('fw_id'), (
                    fw_constants.RESULT_FW_DELETE_INIT))
                ret = self._delete_fw_fab_dev(tenant_id, drvr_name, fw_dict)
        except Exception as exc:
            LOG.error("Exception raised in delete fw %s", str(exc))
        return ret

    def _check_update_fw(self, tenant_id, drvr_name):
        """Update the Firewall config by calling the driver.

        This function calls the device manager routine to update the device
        with modified FW cfg.
        """
        if self.fwid_attr[tenant_id].is_fw_complete():
            fw_dict = self.fwid_attr[tenant_id].get_fw_dict()
            self.modify_fw_device(tenant_id, fw_dict.get('fw_id'), fw_dict)

    def fw_handler_decorator(fw_func):
        @wraps(fw_func)
        def fw_handler_fn(*args):
            try:
                fn_name = fw_func.__name__
                fw_func(*args)
            except Exception as exc:
                LOG.error("Exception in %(name)s %(exc)s",
                          {'name': fn_name, 'exc': str(exc)})

        return fw_handler_fn

    @fw_handler_decorator
    def _fw_create(self, drvr_name, data, cache):
        """Firewall create routine.

        This function updates its local cache with FW parameters.
        It checks if local cache has information about the Policy
        associated with the FW. If not, it means a restart has happened.
        It retrieves the policy associated with the FW by calling
        Openstack API's and calls t he policy create internal routine.
        """

        fw = data.get('firewall')
        tenant_id = fw.get('tenant_id')
        fw_name = fw.get('name')
        fw_id = fw.get('id')
        fw_pol_id = fw.get('firewall_policy_id')
        admin_state = fw.get('admin_state_up')
        rtr_id = None
        if 'router_ids' in fw and len(fw.get('router_ids')) != 0:
            rtr_id = fw.get('router_ids')[0]
        if not admin_state:
            LOG.debug("Admin state disabled")
            return

        name = dfa_dbm.DfaDBMixin.get_project_name(self, tenant_id)
        rtr_name = '_'.join([fw_constants.TENANT_EDGE_RTR, name])

        fw_rtr_name = self.os_helper.get_rtr_name(rtr_id)
        fw_type = None
        if fw_rtr_name == rtr_name:
            fw_type = fw_constants.FW_TENANT_EDGE
        if tenant_id not in self.fwid_attr:
            self.fwid_attr[tenant_id] = FwMapAttr(tenant_id)
        tenant_obj = self.fwid_attr[tenant_id]
        tenant_obj.create_fw(name, fw_pol_id, fw_id, fw_name, fw_type, rtr_id)
        self.tenant_db.store_fw_tenant(fw_id, tenant_id)
        if not cache:
            self._check_create_fw(tenant_id, drvr_name)
        if fw_pol_id is not None and not (
                tenant_obj.is_policy_present(fw_pol_id)):
            pol_data = self.os_helper.get_fw_policy(fw_pol_id)
            if pol_data is not None:
                self.fw_policy_create(pol_data, cache=cache)

    def fw_create(self, data, fw_name=None, cache=False):
        """Top level FW create function. """
        LOG.debug("FW create %s", data)
        try:
            self._fw_create(fw_name, data, cache)
        except Exception as exc:
            LOG.error("Exception in fw_create %s", str(exc))

    @fw_handler_decorator
    def _fw_update(self, drvr_name, data):
        """Update routine for the Firewall.

        Check if FW is already cfgd using the below function
        if self.fwid_attr[tenant_id].is_fw_complete() or
        is_fw_drvr_create_needed():
        The above two functions will take care of whether FW is already
        cfgd or about to be cfgd in case of error.
        If yes, this may be a change in policies attached to FW.
        If no, do a check, create after storing the parameters like
        rtr_id.
        """
        fw = data.get('firewall')
        tenant_id = fw.get('tenant_id')
        if self.fwid_attr[tenant_id].is_fw_complete() or \
           self.fwid_attr[tenant_id].is_fw_drvr_create_needed():
            prev_info_complete = True
        else:
            prev_info_complete = False

        tenant_obj = self.fwid_attr[tenant_id]
        if 'router_ids' in fw and len(fw.get('router_ids')) != 0:
            rtr_id = fw.get('router_ids')[0]
            name = dfa_dbm.DfaDBMixin.get_project_name(self, tenant_id)
            rtr_name = '_'.join([fw_constants.TENANT_EDGE_RTR, name])

            fw_rtr_name = self.os_helper.get_rtr_name(rtr_id)
            fw_type = None
            if fw_rtr_name == rtr_name:
                fw_type = fw_constants.FW_TENANT_EDGE
            tenant_obj.update_fw_params(rtr_id, fw_type)

        if not prev_info_complete:
            self._check_create_fw(tenant_id, drvr_name)

    def fw_update(self, data, fw_name=None):
        """Top level FW update function. """
        LOG.debug("FW Update %s", data)
        self._fw_update(fw_name, data)

    @fw_handler_decorator
    def _fw_delete(self, drvr_name, data):
        """Firewall Delete routine.

        This function calls routines to remove FW from fabric and device.
        It also updates its local cache.
        """
        fw_id = data.get('firewall_id')
        tenant_id = self.tenant_db.get_fw_tenant(fw_id)

        if tenant_id not in self.fwid_attr:
            LOG.error("Invalid tenant id for FW delete %s", tenant_id)
            return

        tenant_obj = self.fwid_attr[tenant_id]
        ret = self._check_delete_fw(tenant_id, drvr_name)
        if ret:
            tenant_obj.delete_fw(fw_id)
            self.tenant_db.del_fw_tenant(fw_id)

    def fw_delete(self, data, fw_name=None):
        """Top level FW delete function. """
        self._fw_delete(fw_name, data)

    def _fw_rule_decode_store(self, data):
        """Misc function to decode the firewall rule from Openstack. """
        fw_rule = data.get('firewall_rule')
        rule = {'protocol': fw_rule.get('protocol'),
                'source_ip_address': fw_rule.get('source_ip_address'),
                'destination_ip_address': fw_rule.get(
                    'destination_ip_address'),
                'source_port': fw_rule.get('source_port'),
                'destination_port': fw_rule.get('destination_port'),
                'action': fw_rule.get('action'),
                'enabled': fw_rule.get('enabled'),
                'name': fw_rule.get('name')}
        return rule

    @fw_handler_decorator
    def _fw_rule_create(self, drvr_name, data, cache):
        """Firewall Rule create routine.

        This function updates its local cache with rule parameters.
        It checks if local cache has information about the Policy
        associated with the rule. If not, it means a restart has happened.
        It retrieves the policy associated with the FW by calling
        Openstack API's and calls t he policy create internal routine.
        """
        tenant_id = data.get('firewall_rule').get('tenant_id')
        fw_rule = data.get('firewall_rule')
        rule = self._fw_rule_decode_store(data)
        fw_pol_id = fw_rule.get('firewall_policy_id')
        rule_id = fw_rule.get('id')
        if tenant_id not in self.fwid_attr:
            self.fwid_attr[tenant_id] = FwMapAttr(tenant_id)
        self.fwid_attr[tenant_id].store_rule(rule_id, rule)
        if not cache:
            self._check_create_fw(tenant_id, drvr_name)
        self.tenant_db.store_rule_tenant(rule_id, tenant_id)
        if fw_pol_id is not None and not (
                self.fwid_attr[tenant_id].is_policy_present(fw_pol_id)):
            pol_data = self.os_helper.get_fw_policy(fw_pol_id)
            if pol_data is not None:
                self.fw_policy_create(pol_data, cache=cache)

    def fw_rule_create(self, data, fw_name=None, cache=False):
        """Top level rule creation routine. """
        LOG.debug("FW Rule create %s", data)
        self._fw_rule_create(fw_name, data, cache)

    @fw_handler_decorator
    def _fw_rule_delete(self, drvr_name, data):
        """Function that updates its local cache after a rule is deleted. """
        rule_id = data.get('firewall_rule_id')
        tenant_id = self.tenant_db.get_rule_tenant(rule_id)

        if tenant_id not in self.fwid_attr:
            LOG.error("Invalid tenant id for FW delete %s", tenant_id)
            return
        tenant_obj = self.fwid_attr[tenant_id]
        # Guess actual FW/policy need not be deleted if this is the active
        # rule, Openstack does not allow it to be deleted
        tenant_obj.delete_rule(rule_id)
        self.tenant_db.del_rule_tenant(rule_id)

    def fw_rule_delete(self, data, fw_name=None):
        """Top level rule delete function. """
        LOG.debug("FW Rule delete %s", data)
        self._fw_rule_delete(fw_name, data)

    @fw_handler_decorator
    def _fw_rule_update(self, drvr_name, data):
        """Firewall Rule update routine.

        Function to decode the updated rules and call routines that
        in turn calls the device routines to update rules.
        """
        LOG.debug("FW Update %s", data)
        tenant_id = data.get('firewall_rule').get('tenant_id')
        fw_rule = data.get('firewall_rule')
        rule = self._fw_rule_decode_store(data)
        rule_id = fw_rule.get('id')
        if tenant_id not in self.fwid_attr or not (
           self.fwid_attr[tenant_id].is_rule_present(rule_id)):
            LOG.error("Incorrect update info for tenant %s", tenant_id)
            return
        self.fwid_attr[tenant_id].rule_update(rule_id, rule)
        self._check_update_fw(tenant_id, drvr_name)

    def fw_rule_update(self, data, fw_name=None):
        """Top level rule update routine. """
        LOG.debug("FW Update Debug")
        self._fw_rule_update(fw_name, data)

    @fw_handler_decorator
    def _fw_policy_delete(self, drvr_name, data):
        """Routine to delete the policy from local cache. """
        policy_id = data.get('firewall_policy_id')
        tenant_id = self.tenant_db.get_policy_tenant(policy_id)

        if tenant_id not in self.fwid_attr:
            LOG.error("Invalid tenant id for FW delete %s", tenant_id)
            return
        tenant_obj = self.fwid_attr[tenant_id]
        # Guess actual FW need not be deleted since if this is the active
        # policy, Openstack does not allow it to be deleted
        tenant_obj.delete_policy(policy_id)
        self.tenant_db.del_policy_tenant(policy_id)

    def fw_policy_delete(self, data, fw_name=None):
        """Top level policy delete routine. """
        LOG.debug("FW Policy Debug")
        self._fw_policy_delete(fw_name, data)

    @fw_handler_decorator
    def _fw_policy_create(self, drvr_name, data, cache):
        """Firewall Policy create routine.

        This function updates its local cache with policy parameters.
        It checks if local cache has information about the rules
        associated with the policy. If not, it means a restart has
        happened. It retrieves the rules associated with the policy by
        calling Openstack API's and calls the rule create internal routine.
        """
        policy = {}
        fw_policy = data.get('firewall_policy')
        tenant_id = fw_policy.get('tenant_id')
        LOG.info("Creating policy for tenant %s", tenant_id)
        policy_id = fw_policy.get('id')
        policy_name = fw_policy.get('name')
        pol_rule_dict = fw_policy.get('firewall_rules')
        if tenant_id not in self.fwid_attr:
            self.fwid_attr[tenant_id] = FwMapAttr(tenant_id)
        policy['name'] = policy_name
        policy['rule_dict'] = pol_rule_dict
        self.fwid_attr[tenant_id].store_policy(policy_id, policy)
        if not cache:
            self._check_create_fw(tenant_id, drvr_name)
        self.tenant_db.store_policy_tenant(policy_id, tenant_id)
        for rule in pol_rule_dict:
            rule_id = rule
            if not self.fwid_attr[tenant_id].is_rule_present(rule_id):
                rule_data = self.os_helper.get_fw_rule(rule_id)
                if rule_data is not None:
                    self.fw_rule_create(rule_data, cache=cache)

    def fw_policy_create(self, data, fw_name=None, cache=False):
        """Top level policy create routine. """
        LOG.debug("FW Policy Debug")
        self._fw_policy_create(fw_name, data, cache)

    def convert_fwdb_event_msg(self, rule, tenant_id, rule_id, policy_id):
        """Convert the Firewall DB to a event message format.

        From inputs from DB, this will create a FW rule dictionary that
        resembles the actual data from Openstack when a rule is created.
        This is usually called after restart, in order to populate local
        cache.
        """
        rule.update({'tenant_id': tenant_id, 'id': rule_id,
                     'firewall_policy_id': policy_id})
        fw_rule_data = {'firewall_rule': rule}
        return fw_rule_data

    def convert_fwdb(self, tenant_id, name, policy_id, fw_id):
        """Convert the Firewall DB to a query response.

        From FWDB inputs, this will create a FW message that resembles the
        actual data from Openstack, when a query for FW is done.
        """
        fw_dict = {'tenant_id': tenant_id, 'name': name, 'id': fw_id,
                   'firewall_policy_id': policy_id,
                   'admin_state_up': True}
        fw_data = {'firewall': fw_dict}
        return fw_data

    def populate_local_cache(self):
        """This populates the local cache after reading the Database.

        It calls the appropriate rule create, fw create routines.
        It doesn't actually call the routine to prepare the fabric or cfg the
        device since it will be handled by retry module.
        """
        fw_dict = self.get_all_fw_db()
        LOG.info("Populating FW Mgr Local Cache")
        for fw_id in fw_dict:
            fw_data = fw_dict.get(fw_id)
            tenant_id = fw_data.get('tenant_id')
            rule_dict = fw_data.get('rules').get('rules')
            policy_id = fw_data.get('rules').get('firewall_policy_id')
            for rule in rule_dict:
                fw_evt_data = self.convert_fwdb_event_msg(rule_dict.get(rule),
                                                          tenant_id, rule,
                                                          policy_id)
                LOG.info("Populating Rules for tenant %s", tenant_id)
                self.fw_rule_create(fw_evt_data, cache=True)
            fw_os_data = self.os_helper.get_fw(fw_id)
            # If enabler is stopped and FW is deleted, then the above routine
            # will fail.
            if fw_os_data is None:
                fw_os_data = self.convert_fwdb(tenant_id, fw_data.get('name'),
                                               policy_id, fw_id)
            LOG.info("Populating FW for tenant %s", tenant_id)
            self.fw_create(fw_os_data, cache=True)
            if fw_data.get('device_status') == 'SUCCESS':
                self.fwid_attr[tenant_id].fw_drvr_created(True)
            else:
                self.fwid_attr[tenant_id].fw_drvr_created(False)
        return fw_dict

    def retry_failure_fab_dev_create(self, tenant_id, fw_data, fw_dict):
        """This module calls routine in fabric to retry the failure cases.

        If device is not successfully cfg/uncfg, it calls the device manager
        routine to cfg/uncfg the device.
        """
        result = fw_data.get('result').split('(')[0]
        is_fw_virt = self.is_device_virtual()
        # Fabric portion
        if result == fw_constants.RESULT_FW_CREATE_INIT:
            name = dfa_dbm.DfaDBMixin.get_project_name(self, tenant_id)
            ret = self.fabric.retry_failure(tenant_id, name, fw_dict,
                                            is_fw_virt, result)
            if not ret:
                LOG.error("Retry failure returned fail for tenant %s",
                          tenant_id)
                return
            else:
                result = fw_constants.RESULT_FW_CREATE_DONE
                self.update_fw_db_final_result(fw_dict.get('fw_id'), result)
        # Device portion
        if result == fw_constants.RESULT_FW_CREATE_DONE:
            if fw_data.get('device_status') != 'SUCCESS':
                ret = self.create_fw_device(tenant_id, fw_dict.get('fw_id'),
                                            fw_dict)
                if ret:
                    self.fwid_attr[tenant_id].fw_drvr_created(True)
                    self.update_fw_db_dev_status(fw_dict.get('fw_id'),
                                                 'SUCCESS')
                    LOG.info("Retry failue return success for create"
                             " tenant %s", tenant_id)

    def retry_failure_fab_dev_delete(self, tenant_id, fw_data, fw_dict):
        """Retry the failure cases for delete.

        This module calls routine in fabric to retry the failure cases for
        delete.
        If device is not successfully cfg/uncfg, it calls the device manager
        routine to cfg/uncfg the device.
        """
        result = fw_data.get('result').split('(')[0]
        name = dfa_dbm.DfaDBMixin.get_project_name(self, tenant_id)
        fw_dict['tenant_name'] = name
        is_fw_virt = self.is_device_virtual()
        if result == fw_constants.RESULT_FW_DELETE_INIT:
            if self.fwid_attr[tenant_id].is_fw_drvr_created():
                ret = self.delete_fw_device(tenant_id, fw_dict.get('fw_id'),
                                            fw_dict)
                if ret:
                    # Device portion
                    self.update_fw_db_dev_status(fw_dict.get('fw_id'),
                                                 '')
                    self.fwid_attr[tenant_id].fw_drvr_created(False)
                    LOG.info("Retry failue dev return success for delete"
                             " tenant %s", tenant_id)
                else:
                    return
            name = dfa_dbm.DfaDBMixin.get_project_name(self, tenant_id)
            ret = self.fabric.retry_failure(tenant_id, name, fw_dict,
                                            is_fw_virt, result)
            if not ret:
                LOG.error("Retry failure returned fail for tenant %s",
                          tenant_id)
                return
            result = fw_constants.RESULT_FW_DELETE_DONE
            self.update_fw_db_final_result(fw_dict.get('fw_id'), result)
            self.delete_fw(fw_dict.get('fw_id'))
            self.fwid_attr[tenant_id].delete_fw(fw_dict.get('fw_id'))
            self.tenant_db.del_fw_tenant(fw_dict.get('fw_id'))

    def fw_retry_failures_create(self):
        """This module is called for retrying the create cases. """
        for tenant_id in self.fwid_attr:
            try:
                with self.fwid_attr[tenant_id].mutex_lock:
                    if self.fwid_attr[tenant_id].is_fw_drvr_create_needed():
                        fw_dict = self.fwid_attr[tenant_id].get_fw_dict()
                        if fw_dict:
                            fw_obj, fw_data = self.get_fw(fw_dict.get('fw_id'))
                            self.retry_failure_fab_dev_create(tenant_id,
                                                              fw_data,
                                                              fw_dict)
                        else:
                            LOG.error("FW data not found for tenant %s",
                                      tenant_id)
            except Exception as exc:
                LOG.error("Exception in retry failure create %s",
                          str(exc))

    def fill_fw_dict_from_db(self, fw_data):
        """
        This routine is called to create a local fw_dict with data from DB.
        """
        rule_dict = fw_data.get('rules').get('rules')
        fw_dict = {'fw_id': fw_data.get('fw_id'),
                   'fw_name': fw_data.get('name'),
                   'firewall_policy_id': fw_data.get('firewall_policy_id'),
                   'fw_type': fw_data.get('fw_type'),
                   'router_id': fw_data.get('router_id'), 'rules': {}}
        for rule in rule_dict:
            fw_dict['rules'][rule] = rule_dict.get(rule)
        return fw_dict

    def fw_retry_failures_delete(self):
        """This routine is called for retrying the delete cases. """
        for tenant_id in self.fwid_attr:
            try:
                with self.fwid_attr[tenant_id].mutex_lock:
                    # For both create and delete case
                    fw_data = self.get_fw_by_tenant_id(tenant_id)
                    if fw_data is None:
                        LOG.info("No FW for tenant %s", tenant_id)
                        continue
                    result = fw_data.get('result').split('(')[0]
                    if result == fw_constants.RESULT_FW_DELETE_INIT:
                        fw_dict = self.fwid_attr[tenant_id].get_fw_dict()
                        # This means a restart has happened before the FW is
                        # completely deleted
                        if not fw_dict:
                            # Need to fill fw_dict from fw_data
                            fw_dict = self.fill_fw_dict_from_db(fw_data)
                        self.retry_failure_fab_dev_delete(tenant_id, fw_data,
                                                          fw_dict)
            except Exception as exc:
                LOG.error("Exception in retry failure delete %s",
                          str(exc))

    def fw_retry_failures(self):
        """Top level retry routine called. """
        if not self.fw_init:
            return
        try:
            self.fw_retry_failures_create()
            self.fw_retry_failures_delete()
        except Exception as exc:
            LOG.error("Exception in retry failures %s", str(exc))



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function

from networking_cisco._i18n import _

from neutronclient.common import extension
from neutronclient.neutron import v2_0 as neutronV20


HOSTING_DEVICE = 'hosting_device'
CFG_AGENT_HOSTING_DEVICES = '/cfg-agent-hosting-devices'


class HostingDeviceHandledByConfigAgent(extension.NeutronClientExtension):
    resource = HOSTING_DEVICE
    resource_plural = '%ss' % resource
    object_path = '/dev_mgr/%s' % resource_plural
    resource_path = '/dev_mgr/%s/%%s' % resource_plural
    versions = ['2.0']
    allow_names = True


class HostingDeviceAssociateWithConfigAgent(extension.ClientExtensionCreate,
                                            HostingDeviceHandledByConfigAgent):

    shell_command = 'cisco-config-agent-associate-hosting-device'

    def get_parser(self, prog_name):
        parser = super(HostingDeviceAssociateWithConfigAgent, self).get_parser(
            prog_name)
        parser.add_argument(
            'config_agent_id',
            help=_('Id of the Cisco configuration agent.'))
        parser.add_argument(
            'hosting_device',
            help=_('Name or id of hosting device to associate.'))
        return parser

    def execute(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        neutron_client = self.get_client()
        neutron_client.format = parsed_args.request_format
        _id_hd = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'hosting_device', parsed_args.hosting_device)
        self.associate_hosting_device_with_config_agent(
            neutron_client, parsed_args.config_agent_id,
            {'hosting_device_id': _id_hd})
        print(_('Associated hosting device \'%(hd)s\' with Cisco '
                'configuration agent \'%(agent)s\'') % {
            'hd': parsed_args.hosting_device,
            'agent': parsed_args.config_agent_id}, file=self.app.stdout,
            end='')
        return [], []

    def associate_hosting_device_with_config_agent(
            self, client, config_agent_id, body):
        """Associates a hosting_device with a config agent."""
        return client.post((ConfigAgentHandlingHostingDevice.resource_path +
                            CFG_AGENT_HOSTING_DEVICES) % config_agent_id,
                           body=body)


class HostingDeviceDisassociateFromConfigAgent(
        extension.ClientExtensionCreate, HostingDeviceHandledByConfigAgent):

    shell_command = 'cisco-config-agent-disassociate-hosting-device'

    def get_parser(self, prog_name):
        parser = super(HostingDeviceDisassociateFromConfigAgent,
                       self).get_parser(prog_name)
        parser.add_argument(
            'config_agent_id',
            help=_('Id of the Cisco configuration agent.'))
        parser.add_argument(
            'hosting_device',
            help=_('Name or id of hosting device to disassociate.'))
        return parser

    def execute(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        neutron_client = self.get_client()
        neutron_client.format = parsed_args.request_format
        _id_hd = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'hosting_device', parsed_args.hosting_device)
        self.disassociate_hosting_device_with_config_agent(
            neutron_client, parsed_args.config_agent_id, _id_hd)
        print(_('Disassociated hosting device \'%(hd)s\' from Cisco '
                'configuration agent \'%(agent)s\'') % {
            'hd': parsed_args.hosting_device,
            'agent': parsed_args.config_agent_id}, file=self.app.stdout,
            end='')
        return [], []

    def disassociate_hosting_device_with_config_agent(
            self, client, config_agent_id, hosting_device_id):
        """Disassociates a hosting_device with a config agent."""
        return client.delete((ConfigAgentHandlingHostingDevice.resource_path +
                              CFG_AGENT_HOSTING_DEVICES + "/%s") % (
            config_agent_id, hosting_device_id))


class HostingDeviceHandledByConfigAgentList(extension.ClientExtensionList,
                                            HostingDeviceHandledByConfigAgent):
    shell_command = 'cisco-config-agent-list-hosting-devices'
    list_columns = ['id', 'name', 'admin_state_up', 'template_id']

    def get_parser(self, prog_name):
        parser = super(HostingDeviceHandledByConfigAgentList,
                       self).get_parser(prog_name)
        parser.add_argument(
            'config_agent_id',
            help=_('Id of the Cisco configuration agent to query.'))
        return parser

    def call_server(self, neutron_client, search_opts, parsed_args):
        data = self.list_hosting_device_handled_by_config_agent(
            neutron_client, parsed_args.config_agent_id, **search_opts)
        return data

    def list_hosting_device_handled_by_config_agent(
            self, client, cfg_agent_id, **_params):
        """Fetches a list of hosting devices handled by a config agent."""
        return client.get((ConfigAgentHandlingHostingDevice.resource_path +
                           CFG_AGENT_HOSTING_DEVICES) % cfg_agent_id,
                          params=_params)


AGENT = 'agent'
HOSTING_DEVICE_CFG_AGENTS = '/hosting-device-cfg-agents'


class ConfigAgentHandlingHostingDevice(extension.NeutronClientExtension):
    resource = AGENT
    resource_plural = '%ss' % resource
    object_path = '/%s' % resource_plural
    resource_path = '/%s/%%s' % resource_plural
    versions = ['2.0']
    allow_names = True


class ConfigAgentHandlingHostingDeviceList(extension.ClientExtensionList,
                                           ConfigAgentHandlingHostingDevice):

    shell_command = 'cisco-hosting-device-list-config-agents'
    list_columns = ['id', 'alive', 'agent_type', 'admin_state_up', 'host']

    def extend_list(self, data, parsed_args):
        for agent in data:
            agent['alive'] = ":-)" if agent.get('alive') else 'xxx'

    def get_parser(self, prog_name):
        parser = super(ConfigAgentHandlingHostingDeviceList, self).get_parser(
            prog_name)
        parser.add_argument('hosting_device',
                            help=_('Name or id of hosting device to query.'))
        return parser

    def call_server(self, neutron_client, search_opts, parsed_args):
        _id = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'hosting_device', parsed_args.hosting_device)
        data = self.list_config_agents_handling_hosting_device(
            neutron_client, _id, **search_opts)
        return data

    def list_config_agents_handling_hosting_device(
            self, client, hosting_device_id, **_params):
        """Fetches a list of config agents handling a hosting device."""
        resource_path = '/dev_mgr/hosting_devices/%s'
        return client.get((resource_path + HOSTING_DEVICE_CFG_AGENTS) %
                          hosting_device_id, params=_params)



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import netaddr

from oslo_config import cfg

from networking_cisco._i18n import _

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent import cfg_exceptions as cfg_exc
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_cfg_syncer)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_snippets)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.iosxe import (
    cisco_iosxe_snippets as snippets)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.iosxe import (
    iosxe_routing_driver as iosxe_driver)
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole

LOG = logging.getLogger(__name__)


DEVICE_OWNER_ROUTER_GW = bc.constants.DEVICE_OWNER_ROUTER_GW
ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR
ROUTER_ROLE_HA_REDUNDANCY = cisco_constants.ROUTER_ROLE_HA_REDUNDANCY
ROUTER_ROLE_GLOBAL = cisco_constants.ROUTER_ROLE_GLOBAL

ASR1K_DRIVER_OPTS = [
    cfg.BoolOpt('enable_multi_region',
                default=False,
                help=_("If enabled, the agent will maintain a heartbeat "
                       "against its hosting devices. If a device dies and "
                       "recovers, the agent will then trigger a configuration "
                       "resync.")),
    cfg.StrOpt('region_id',
               default='L3FR001',
               help=_("Label to use for this deployments region-id")),
    cfg.ListOpt('other_region_ids',
                default=['L3FR002', 'L3FR003'],
                help=_("Label for other region-ids")),
]

cfg.CONF.register_opts(ASR1K_DRIVER_OPTS, "multi_region")


class ASR1kRoutingDriver(iosxe_driver.IosXeRoutingDriver):

    def __init__(self, **device_params):
        super(ASR1kRoutingDriver, self).__init__(**device_params)
        self._fullsync = False
        self._deployment_id = "zxy"
        self.hosting_device = {'id': device_params.get('id'),
                               'device_id': device_params.get('device_id')}

    # ============== Public functions ==============
    def send_empty_cfg(self):
        LOG.debug("send empty config")
        conf_str = asr1k_snippets.EMPTY_SNIPPET
        self._edit_running_config(conf_str, 'EMPTY_SNIPPET')

    def internal_network_added(self, ri, port):
        if self._is_port_v6(port):
            LOG.debug("Adding IPv6 internal network port: %(port)s for router "
                      "%(r_id)s", {'port': port, 'r_id': ri.id})
            gw_ip = self._get_item(
                port['subnets'], port['ip_info']['subnet_id'],
                'id')['gateway_ip']
            self._create_sub_interface_v6(ri, port, False, gw_ip)
        else:
            # IPv4 handling
            if self._is_global_router(ri):
                # The global router is modeled as the default vrf in the ASR1k.
                # When an external gateway is configured, a normal "internal"
                # interface is created in the default vrf that is in the same
                # subnet as the ext-net.
                LOG.debug("Attaching Global router %(r_id)s to external "
                          "network using port %(p_id)s", {'r_id': ri.id,
                                                          'p_id': port['id']})
                self.external_gateway_added(ri, port)
            else:
                LOG.debug("Adding IPv4 internal network port: %(p_id)s "
                          "for router %(r_id)s", {'p_id': port['id'],
                                                  'r_id': ri.id})
                details = port['change_details']
                LOG.debug('Add - change details: former: %(f_ips)s, old: '
                          '%(o_ips)s, new: %(n_ips)s, current: %(c_ips)s',
                          {'f_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                     details['former_ports']],
                           'o_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                     details['old_ports']],
                           'n_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                     details['new_ports']],
                           'c_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                     details['current_ports']]})
                former_ports = details['former_ports']
                old_ports = details['old_ports']
                if port['ip_info']['is_primary'] is False:
                    # port to be configured as a secondary
                    LOG.debug('Configuring %(ip)s as SECONDARY address for '
                              'port %(p_id)s of router %(r_id)s',
                              {'ip': port['fixed_ips'][0]['ip_address'],
                               'p_id': port['id'], 'r_id': ri.id})
                    self._set_secondary_ipv4(ri, port)
                elif not former_ports:
                    # port is first one on network so must create sub-interface
                    # and configure port as primary
                    LOG.debug('Creating sub-interface and configuring %(ip)s '
                              'as PRIMARY address for first port %(p_id)s on '
                              'network for router %(r_id)s',
                              {'ip': port['fixed_ips'][0]['ip_address'],
                               'p_id': port['id'], 'r_id': ri.id})
                    self._create_sub_interface(ri, port)
                elif old_ports and old_ports[0] == former_ports[0]['id']:
                    # former primary will be deleted so sub-interface already
                    # exists so can just configure port as primary
                    LOG.debug('Configuring %(ip)s as PRIMARY address for port '
                              '%(p_id)s of router %(r_id)s since port '
                              '%(fp_id)s providing former primary address '
                              '%(f_ip)s is to be deleted',
                              {'ip': port['fixed_ips'][0]['ip_address'],
                               'p_id': port['id'], 'r_id': ri.id,
                               'fp_id': former_ports[0]['id'],
                               'f_ip': former_ports[0]['fixed_ips'][0][
                                   'ip_address']})
                    self._set_primary_ipv4(ri, port)
                else:
                    # port is the new primary
                    LOG.debug('Configuring %(ip)s as new PRIMARY address for '
                              'port %(p_id)s of router %(r_id)s',
                              {'ip': port['fixed_ips'][0]['ip_address'],
                               'p_id': port['id'], 'r_id': ri.id})
                    self._set_primary_ipv4(ri, port)
                    # former primary remains and thus must become a secondary
                    LOG.debug('Configuring %(ip)s as SECONDARY address for '
                              'former primary port %(p_id)s of router '
                              '%(r_id)s',
                              {'ip': former_ports[0]['fixed_ips'][0][
                                  'ip_address'],
                               'p_id': former_ports[0]['id'], 'r_id': ri.id})
                    self._set_subnet_info(
                        former_ports[0],
                        former_ports[0]['fixed_ips'][0]['subnet_id'], False)
                    self._set_secondary_ipv4(ri, former_ports[0])

    def internal_network_removed(self, ri, port):
        if self._is_global_router(ri):
            self._remove_sub_interface(port)
        else:
            details = port['change_details']
            LOG.debug('Remove - change details: former: %(f_ips)s, old: '
                      '%(o_ips)s, new: %(n_ips)s, current: %(c_ips)s',
                      {'f_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                 details['former_ports']],
                       'o_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                 details['old_ports']],
                       'n_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                 details['new_ports']],
                       'c_ips': [p['fixed_ips'][0]['ip_address'] for p in
                                 details['current_ports']]})
            former_ports = details['former_ports']
            new_ports = details['new_ports']
            current_ports = details['current_ports']
            if port['ip_info']['is_primary'] is False:
                LOG.debug('Port %(p_id)s that provides %(ip)s used as '
                          'SECONDARY address is to be removed',
                          {'p_id': port['id'],
                           'ip': port['fixed_ips'][0]['ip_address']})
                # since port is a secondary it can just be removed
                if current_ports:
                    # only explicitly de-configure if other ports on same
                    # network remain since the sub-interface removal will
                    # otherwise take care of secondaries automatically
                    LOG.debug('IPv4 address %(ip)s of port %(p_id)s is '
                              'DECONFIGURED as SECONDARY address since router '
                              '%(r_id)s has other ports on the network',
                              {'ip': port['fixed_ips'][0]['ip_address'],
                               'p_id': port['id'], 'r_id': ri.id})
                    self._remove_secondary_ipv4(ri, port)
            elif len(former_ports) == 1:
                # port was the only port on network (thus also the primary)
                if not new_ports:
                    # no new port on that network is added so we can just
                    # remove the sub-interface
                    LOG.debug('Port %(p_id)s that provides %(ip)s used as '
                              'PRIMARY address is to be removed and router '
                              '%(r_id)s has NO new ports NOR remaining ports '
                              'on network so SUB-INTERFACE WILL BE DELETED.',
                              {'p_id': port['id'],
                               'ip': port['fixed_ips'][0]['ip_address'],
                               'r_id': ri.id})
                    self._remove_sub_interface(port)
                else:
                    # the new primary has already be configured as new ports
                    # are processed before ports that have been removed
                    # NOTHING MORE TO DD
                    LOG.debug('The router (%(r_id)s)\'s ONLY port %(p_id)s on '
                              'network that thus provided %(ip)s used as '
                              'PRIMARY address is to be removed. But '
                              'the router has a NEW port %(np_id)s on network '
                              'that with IPv4 %(n_ip)s has already been '
                              'configured as primary so NOTHING MORE TO DO '
                              'HERE.',
                              {'r_id': ri.id, 'p_id': port['id'],
                               'ip': port['fixed_ips'][0]['ip_address'],
                               'np_id': current_ports[0]['id'],
                               'n_ip': current_ports[0]['fixed_ips'][0][
                                   'ip_address']})
            elif not new_ports:
                # port was one of several ports on network but no new ports
                # were added so one of the existing ports (if any exist) will
                # change role from secondary to be the new primary
                if current_ports:
                    LOG.debug('Port %(p_id)s that provides %(ip)s used as '
                              'PRIMARY address is to be removed. But router '
                              '%(r_id)s has other REMAINING port %(fp_id)s '
                              'with IPv4 address %(f_ip)s that is '
                              'RECONFIGURED from secondary to primary.',
                              {'p_id': port['id'],
                               'ip': port['fixed_ips'][0]['ip_address'],
                               'r_id': ri.id,
                               'fp_id': current_ports[0]['id'],
                               'f_ip': current_ports[0]['fixed_ips'][0][
                                   'ip_address']})
                    self._set_subnet_info(
                        current_ports[0],
                        current_ports[0]['fixed_ips'][0]['subnet_id'], False)
                    self._remove_secondary_ipv4(ri, current_ports[0])
                    current_ports[0]['ip_info']['is_primary'] = True
                    self._set_primary_ipv4(ri, current_ports[0])
                else:
                    # no ports on that network remain so we can just remove
                    # the sub-interface
                    LOG.debug('Port %(p_id)s that provides %(ip)s used as '
                              'PRIMARY address is to be removed. Router '
                              '%(r_id)s has NO remaining ports NOR any new '
                              'ports on network so SUB-INTERFACE WILL BE '
                              'DELETED.',
                              {'p_id': port['id'],
                               'ip': port['fixed_ips'][0]['ip_address'],
                               'r_id': ri.id})
                    self._remove_sub_interface(port)
            elif new_ports[0]['id'] == current_ports[0]['id']:
                # one of the new ports is the new primary at it will already
                # have overwritten the old primary so we're done
                # NOTHING MORE TO DO
                LOG.debug('Port %(p_id)s that provides %(ip)s used as '
                          'PRIMARY address is to be removed. But router '
                          '%(r_id)s has a NEW port %(np_id)s with IPv4 '
                          'address %(n_ip)s that has already been'
                          'CONFIGURED AS PRIMARY ',
                          {'p_id': port['id'],
                           'ip': port['fixed_ips'][0]['ip_address'],
                           'r_id': ri.id,
                           'np_id': current_ports[0]['id'],
                           'n_ip': current_ports[0]['fixed_ips'][0][
                               'ip_address']})
            else:
                # no new new port is a new primary so one of the existing
                # ports will change role from secondary to be the new primary
                LOG.debug('Port %(p_id)s that provides %(ip)s used as '
                          'PRIMARY address is to be removed. But router '
                          '%(r_id)s has new ports and a REMAINING port '
                          '%(fp_id)s with IPv4 address %(f_ip)s that is '
                          'RECONFIGURED FROM SECONDARY TO PRIMARY.',
                          {'p_id': port['id'],
                           'ip': port['fixed_ips'][0]['ip_address'],
                           'r_id': ri.id,
                           'fp_id': current_ports[0]['id'],
                           'f_ip': current_ports[0]['fixed_ips'][0][
                               'ip_address']})
                self._set_subnet_info(
                    current_ports[0],
                    current_ports[0]['fixed_ips'][0]['subnet_id'], False)
                self._remove_secondary_ipv4(ri, current_ports[0])
                current_ports[0]['ip_info']['is_primary'] = True
                self._set_primary_ipv4(ri, current_ports[0])

    def external_gateway_added(self, ri, ext_gw_port):
        # global router handles IP assignment, HSRP setup
        # tenant router handles interface creation and default route
        # within VRFs
        if self._is_global_router(ri):
            self._handle_external_gateway_added_global_router(ri, ext_gw_port)
        else:
            self._handle_external_gateway_added_normal_router(ri, ext_gw_port)

    def external_gateway_removed(self, ri, ext_gw_port):
        if self._is_global_router(ri):
            self._remove_sub_interface(ext_gw_port)
        else:
            ex_gw_ip = self._get_item(ext_gw_port['subnets'],
                                      ext_gw_port['ip_info']['subnet_id'],
                                      'id')['gateway_ip']
            if (ex_gw_ip and
                    ext_gw_port['device_owner'] == DEVICE_OWNER_ROUTER_GW):
                # Remove default route via this network's gateway ip
                if self._is_port_v6(ext_gw_port):
                    self._remove_default_route_v6(ri, ex_gw_ip, ext_gw_port)
                else:
                    self._set_nat_pool(ri, ext_gw_port, True)
                    self._remove_default_route(ri, ext_gw_port)

    def floating_ip_added(self, ri, ext_gw_port, floating_ip, fixed_ip):
        self._add_floating_ip_asr1k(ri, ext_gw_port, floating_ip, fixed_ip)

    def floating_ip_removed(self, ri, ext_gw_port, floating_ip, fixed_ip):
        self._remove_floating_ip(ri, ext_gw_port, floating_ip, fixed_ip)

    def disable_internal_network_NAT(self, ri, port, ext_gw_port,
                                     itfc_deleted=False):
        self._remove_internal_nw_nat_rules(ri, [port], ext_gw_port,
                                           itfc_deleted)

    def enable_router_interface(self, ri, port):
        # Enable the router interface
        interface = self._get_interface_name_from_hosting_port(port)
        self._enable_sub_interface(interface)

    def disable_router_interface(self, ri, port=None):
        # Disable the router interface
        if not port:
            ex_gw_port = ri.router.get('gw_port', None)
            if ex_gw_port:
                ext_interface = (self._get_interface_name_from_hosting_port(
                                 ex_gw_port))
                self._disable_sub_interface(ext_interface)
            internal_ports = ri.router.get(bc.constants.INTERFACE_KEY, [])
            for port in internal_ports:
                internal_interface = (
                        self._get_interface_name_from_hosting_port(port))
                self._disable_sub_interface(internal_interface)
        else:
            interface = self._get_interface_name_from_hosting_port(port)
            self._disable_sub_interface(interface)

    def cleanup_invalid_cfg(self, hd, routers):
        cfg_syncer = asr1k_cfg_syncer.ConfigSyncer(routers, self, hd)
        cfg_syncer.delete_invalid_cfg()

    def get_configuration(self):
        return self._get_running_config(split=False)

    # ============== Internal "preparation" functions  ==============
    def _get_vrf_name(self, ri):
        """
        overloaded method for generating a vrf_name that supports
        region_id
        """
        router_id = ri.router_name()[:self.DEV_NAME_LEN]
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region

        if is_multi_region_enabled:
            region_id = cfg.CONF.multi_region.region_id
            vrf_name = "%s-%s" % (router_id, region_id)
        else:
            vrf_name = router_id
        return vrf_name

    def _get_acl_name_from_vlan(self, vlan):
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region

        if is_multi_region_enabled:
            region_id = cfg.CONF.multi_region.region_id
            acl_name = "neutron_acl_%s_%s" % (region_id, vlan)
        else:
            acl_name = "neutron_acl_%s" % vlan
        return acl_name

    def _generate_acl_num_from_port(self, port):
        port_id = port['id'][:8]  # Taking only the first 8 chars
        vlan = self._get_interface_vlan_from_hosting_port(port)
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region

        if is_multi_region_enabled:
            region_id = cfg.CONF.multi_region.region_id
            acl_name = "neutron_acl_%s_%s_%s" % (region_id, vlan, port_id)
        else:
            acl_name = "neutron_acl_%s_%s" % (vlan, port_id)
        return acl_name

    def _get_interface_name_from_hosting_port(self, port):
        """
        Extract the underlying subinterface name for a port
        e.g. Port-channel10.200 or GigabitEthernet0/0/0.500
        """
        try:
            vlan = port['hosting_info']['segmentation_id']
            int_prefix = port['hosting_info']['physical_interface']
            return '%s.%s' % (int_prefix, vlan)
        except KeyError as e:
            params = {'key': e}
            raise cfg_exc.DriverExpectedKeyNotSetException(**params)

    @staticmethod
    def _set_subnet_info(port, subnet_id, is_primary=True):
        ip = next((i['ip_address'] for i in port['fixed_ips']
                   if i['subnet_id'] == subnet_id), None)
        if ip is None:
            # there will be KeyError exception later if this happens
            LOG.error('Port %(p_id)s lacks IP address on subnet %(s_id)s',
                      {'p_id': port['id'], 's_id': subnet_id})
            return
        subnet = next(sn for sn in port['subnets'] if sn['id'] == subnet_id)
        prefixlen = netaddr.IPNetwork(subnet['cidr']).prefixlen
        port['ip_info'] = {'subnet_id': subnet_id, 'is_primary': is_primary,
                           'ip_cidr': "%s/%s" % (ip, prefixlen)}

    @staticmethod
    def _get_item(list_containing_dicts_entries, attribute_value,
                  attribute_name='subnet_id'):
        """Searches a list of dicts and returns the first matching entry

        The dict entry returned contains the attribute 'attribute_name' whose
        value equals 'attribute_value'. If no such dict is found in the list
        an empty dict is returned.
        """
        for item in list_containing_dicts_entries:
            if item.get(attribute_name) == attribute_value:
                return item
        return {}

    def _enable_itfcs(self, conn):
        """For ASR we don't need to do anything"""
        return True

    def _handle_external_gateway_added_global_router(self, ri, ext_gw_port):
        # TODO(bobmel): Get the HA virtual IP correctly
        # NOTE(sridar): This seems to work fine. Keeping this todo until
        #               more testing.
        subnet_id = ext_gw_port['ip_info']['subnet_id']
        if self._is_port_v6(ext_gw_port):
            LOG.debug("Adding IPv6 external network port: %(port)s on "
                      "subnet: (subnet)s for global router %(r_id)s",
                      {'port': ext_gw_port['id'], 'subnet': subnet_id,
                       'r_id': ri.id})
            virtual_gw_port = ext_gw_port[ha.HA_INFO]['ha_port']
            sub_itfc_ip = self._get_item(virtual_gw_port['fixed_ips'],
                                         subnet_id)['ip_address']
            self._create_sub_interface_v6(ri, ext_gw_port, True, sub_itfc_ip)
        else:
            LOG.debug("Adding IPv4 external network port: %(port)s on "
                      "subnet: (subnet)s for global router %(r_id)s",
                      {'port': ext_gw_port['id'], 'subnet': subnet_id,
                       'r_id': ri.id})
            self._create_external_sub_interface(ri, ext_gw_port)

    def _handle_external_gateway_added_normal_router(self, ri, ext_gw_port):
        # Default routes are mapped to tenant router VRFs. Global Router
        # is not aware of tenant routers with ext network assigned. Thus,
        # default route must be handled per tenant router.
        ex_gw_ip = self._get_item(
            ext_gw_port['subnets'], ext_gw_port['ip_info']['subnet_id'],
            'id')['gateway_ip']

        sub_interface = self._get_interface_name_from_hosting_port(ext_gw_port)
        vlan_id = self._get_interface_vlan_from_hosting_port(ext_gw_port)
        if (self._fullsync and
                int(vlan_id) in self._existing_cfg_dict['interfaces']):
            LOG.debug("Sub-interface already exists, will not create it")
        else:
            LOG.debug("Adding IPv4 external network port: %(port)s for tenant "
                      "router %(r_id)s", {'port': ext_gw_port['id'],
                                          'r_id': ri.id})
            if ri.router['admin_state_up'] and ext_gw_port['admin_state_up']:
                self._enable_sub_interface(sub_interface)
            else:
                self._disable_sub_interface(sub_interface)

        if ex_gw_ip:
            # Set default route via this network's gateway ip
            if self._is_port_v6(ext_gw_port):
                self._add_default_route_v6(ri, ex_gw_ip, ext_gw_port)
            else:
                self._set_nat_pool(ri, ext_gw_port, False)
                self._add_default_route(ri, ext_gw_port)

    def _create_sub_interface(self, ri, port):
        vlan = self._get_interface_vlan_from_hosting_port(port)
        if (self._fullsync and
                int(vlan) in self._existing_cfg_dict['interfaces']):
            LOG.info("Sub-interface already exists, will not create it")
            return
        vrf_name = self._get_vrf_name(ri)
        net_mask = netaddr.IPNetwork(port['ip_info']['ip_cidr']).netmask
        # get port's ip address for the subnet we're processing
        ip = self._get_item(port['fixed_ips'],
                            port['ip_info']['subnet_id'])['ip_address']
        sub_interface = self._get_interface_name_from_hosting_port(port)
        self._do_create_sub_interface(sub_interface, vlan, vrf_name, ip,
                                      net_mask)
        self._conditionally_add_ha_hsrp_asr1k(ri, port)

    def _create_external_sub_interface(self, ri, port):
        vlan = self._get_interface_vlan_from_hosting_port(port)
        if (self._fullsync and
                int(vlan) in self._existing_cfg_dict['interfaces']):
            LOG.info("Sub-interface already exists, will not create it")
            return
        net_mask = netaddr.IPNetwork(port['ip_info']['ip_cidr']).netmask
        # get port's ip address for the subnet we're processing
        ip = self._get_item(port['fixed_ips'],
                            port['ip_info']['subnet_id'])['ip_address']
        sub_interface = self._get_interface_name_from_hosting_port(port)
        if port['ip_info']['is_primary'] is True:
            self._do_create_external_sub_interface(sub_interface, vlan, ip,
                                                   net_mask)
        else:
            # this will only happen for global routers or when a tenant
            # router is connected to multiple subnets on the same network
            self._do_set_secondary(sub_interface, ip, net_mask)
        self._conditionally_add_ha_hsrp_asr1k(ri, port)

    def _set_primary_ipv4(self, ri, port):
        sub_interface = self._get_interface_name_from_hosting_port(port)
        ip = self._get_item(port['fixed_ips'],
                            port['ip_info']['subnet_id'])['ip_address']
        net_mask = netaddr.IPNetwork(port['ip_info']['ip_cidr']).netmask
        self._do_set_primary(sub_interface, ip, net_mask)
        self._conditionally_add_ha_hsrp_asr1k(ri, port)

    def _set_secondary_ipv4(self, ri, port):
        net_mask = netaddr.IPNetwork(
            port['ip_info']['ip_cidr']).netmask
        # get port's ip address for the subnet we're processing
        subnet_id = port['ip_info']['subnet_id']
        ip = self._get_item(port['fixed_ips'], subnet_id)['ip_address']
        sub_interface = self._get_interface_name_from_hosting_port(
            port)
        self._do_set_secondary(sub_interface, ip, net_mask)
        self._conditionally_add_ha_hsrp_asr1k(ri, port)

    def _remove_secondary_ipv4(self, ri, port):
        net_mask = netaddr.IPNetwork(
            port['ip_info']['ip_cidr']).netmask
        # get port's ip address for the subnet we're processing
        subnet_id = port['ip_info']['subnet_id']
        ip = self._get_item(port['fixed_ips'], subnet_id)['ip_address']
        sub_interface = self._get_interface_name_from_hosting_port(
            port)
        self._do_unset_secondary(sub_interface, ip, net_mask)
        self._conditionally_remove_ha_hsrp_asr1k(ri, port, sub_interface)

    def _do_create_sub_interface(self, sub_interface, vlan_id, vrf_name, ip,
                                 mask):
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        if is_multi_region_enabled:
            region_id = cfg.CONF.multi_region.region_id
            conf_str = asr1k_snippets.CREATE_SUBINTERFACE_REGION_ID_WITH_ID % (
                sub_interface, region_id, vlan_id, vrf_name, ip, mask)
        else:
            conf_str = asr1k_snippets.CREATE_SUBINTERFACE_WITH_ID % (
                sub_interface, vlan_id, vrf_name, ip, mask)
        self._edit_running_config(conf_str, 'CREATE_SUBINTERFACE_WITH_ID')

    def _do_create_external_sub_interface(self, sub_interface, vlan_id, ip,
                                          mask):
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        if is_multi_region_enabled:
            region_id = cfg.CONF.multi_region.region_id
            conf_str = (
                asr1k_snippets.CREATE_SUBINTERFACE_EXT_REGION_ID_WITH_ID % (
                    sub_interface, region_id, vlan_id, ip, mask))
        else:
            conf_str = (
                asr1k_snippets.CREATE_SUBINTERFACE_EXTERNAL_WITH_ID % (
                    sub_interface, vlan_id, ip, mask))
        self._edit_running_config(conf_str, 'CREATE_SUBINTERFACE_WITH_ID')

    def _do_set_primary(self, sub_interface, ip, mask):
        conf_str = (
            asr1k_snippets.SET_INTERFACE_PRIMARY_IP % (
                    sub_interface, ip, mask))
        self._edit_running_config(conf_str, 'SET_SUBINTERFACE_PRIMARY_IP')

    def _do_unset_primary(self, sub_interface, ip, mask):
        conf_str = (
            asr1k_snippets.REMOVE_INTERFACE_PRIMARY_IP % (
                sub_interface, ip, mask))
        self._edit_running_config(conf_str, 'REMOVE_SUBINTERFACE_PRIMARY_IP')

    def _do_set_secondary(self, sub_interface, ip, mask):
        conf_str = (
            asr1k_snippets.SET_INTERFACE_SECONDARY_IP % (
                    sub_interface, ip, mask))
        self._edit_running_config(conf_str, 'SET_SUBINTERFACE_SECONDARY_IP')

    def _do_unset_secondary(self, sub_interface, ip, mask):
        conf_str = (
            asr1k_snippets.REMOVE_INTERFACE_SECONDARY_IP % (
                sub_interface, ip, mask))
        self._edit_running_config(conf_str, 'REMOVE_SUBINTERFACE_SECONDARY_IP')

    def _enable_sub_interface(self, sub_interface):
        LOG.debug("Enabling network sub interface: %s", sub_interface)
        conf_str = snippets.ENABLE_INTF % sub_interface
        self._edit_running_config(conf_str, 'ENABLE_INTF')

    def _disable_sub_interface(self, sub_interface):
        LOG.debug("Disabling network sub interface: %s", sub_interface)
        conf_str = snippets.DISABLE_INTF % sub_interface
        self._edit_running_config(conf_str, 'DISABLE_INTF')

    def _set_nat_pool(self, ri, gw_port, is_delete):
        vrf_name = self._get_vrf_name(ri)
        if ri.router.get(ROUTER_ROLE_ATTR) == ROUTER_ROLE_HA_REDUNDANCY:
            the_port = gw_port[ha.HA_INFO]['ha_port']
        else:
            the_port = gw_port
        subnet_id = gw_port['ip_info']['subnet_id']
        fixed_ip = self._get_item(the_port['fixed_ips'], subnet_id)
        pool_ip = fixed_ip['ip_address']
        pool_ip_prefix_len = fixed_ip['prefixlen']
        #TODO(ebobmel) We need to modify the pool name generation if we
        #              will have multiple NAT pools per VRF
        pool_name = "%s_nat_pool" % (vrf_name)
        pool_net = "%s/%s" % (pool_ip, pool_ip_prefix_len)
        pool_net = netaddr.IPNetwork(pool_net)
        if self._fullsync and pool_ip in self._existing_cfg_dict['pools']:
            LOG.info("Pool already exists, skipping")
            return

        try:
            if is_delete:
                conf_str = asr1k_snippets.DELETE_NAT_POOL % (
                    pool_name, pool_ip, pool_ip, pool_net.netmask)
                # TODO(update so that hosting device name is passed down)
                self._edit_running_config(conf_str, 'DELETE_NAT_POOL')
            else:
                conf_str = asr1k_snippets.CREATE_NAT_POOL % (
                    pool_name, pool_ip, pool_ip, pool_net.netmask)
                # TODO(update so that hosting device name is passed down)
                self._edit_running_config(conf_str, 'CREATE_NAT_POOL')
        #except cfg_exc.IOSXEConfigException as cse:
        except Exception as cse:
            LOG.error("Temporary disable NAT_POOL exception handling: %s",
                      cse)

    def _add_default_route(self, ri, ext_gw_port):
        if self._fullsync and (ri.router_id in
                               self._existing_cfg_dict['routes']):
            LOG.debug("Default route already exists, skipping")
            return
        ext_gw_ip = self._get_item(
            ext_gw_port['subnets'], ext_gw_port['ip_info']['subnet_id'],
            'id')['gateway_ip']
        if ext_gw_ip:
            vrf_name = self._get_vrf_name(ri)
            out_itfc = self._get_interface_name_from_hosting_port(ext_gw_port)
            conf_str = asr1k_snippets.SET_DEFAULT_ROUTE_WITH_INTF % (
                vrf_name, out_itfc, ext_gw_ip)
            self._edit_running_config(conf_str, 'SET_DEFAULT_ROUTE_WITH_INTF')

    def _remove_default_route(self, ri, ext_gw_port):
        ext_gw_ip = self._get_item(
            ext_gw_port['subnets'], ext_gw_port['ip_info']['subnet_id'],
            'id')['gateway_ip']
        if ext_gw_ip:
            vrf_name = self._get_vrf_name(ri)
            out_itfc = self._get_interface_name_from_hosting_port(ext_gw_port)
            conf_str = asr1k_snippets.REMOVE_DEFAULT_ROUTE_WITH_INTF % (
                vrf_name, out_itfc, ext_gw_ip)
            self._edit_running_config(conf_str,
                                      'REMOVE_DEFAULT_ROUTE_WITH_INTF')

    def _conditionally_add_ha_hsrp_asr1k(self, ri, port):
        if ri.router.get(ha.ENABLED, False):
            if port.get(ha.HA_INFO) is None:
                # We are missing HA data, candidate for retrying
                params = {'r_id': ri.router_id, 'p_id': port['id'],
                          'port': port}
                raise cfg_exc.HAParamsMissingException(**params)
        else:
            return
        priority = None
        if ri.router.get(ROUTER_ROLE_ATTR) in (ROUTER_ROLE_HA_REDUNDANCY,
                                               ROUTER_ROLE_GLOBAL):
            for router in ri.router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                if ri.router['id'] == router['id']:
                    priority = router[ha.PRIORITY]
        else:
            priority = ri.router[ha.DETAILS][ha.PRIORITY]
        port_ha_info = port[ha.HA_INFO]
        group = port_ha_info['group']
        subnet_id = port['ip_info']['subnet_id']
        vip = self._get_item(port_ha_info['ha_port']['fixed_ips'],
                             subnet_id)['ip_address']
        vlan = port['hosting_info']['segmentation_id']
        if vip and group and priority:
            is_primary = port['ip_info']['is_primary']
            sub_interface = self._get_interface_name_from_hosting_port(port)
            self._do_add_ha_hsrp_asr1k(sub_interface, priority, group, vip,
                                       vlan, is_primary)

    def _conditionally_remove_ha_hsrp_asr1k(self, ri, port, sub_interface):
        if ri.router.get(ha.ENABLED, False):
            if port.get(ha.HA_INFO) is None:
                ip = self._get_item(port['fixed_ips'],
                                    port['ip_info']['subnet_id'])['ip_address']
                LOG.warning(
                    "Could not remove HSRP VIP address for port %(p_id)s "
                    "with interface IP %(addr)s because plugin did not "
                    "provide needed information. Manual removal by admin "
                    "required.", {'p_id': port['id'], 'addr': ip})
                return
        else:
            return
        subnet_id = port['ip_info']['subnet_id']
        is_primary = port['ip_info']['is_primary']
        port_ha_info = port[ha.HA_INFO]
        group = port_ha_info['group']
        hsrp_ip = self._get_item(port_ha_info['ha_port']['fixed_ips'],
                                 subnet_id)['ip_address']
        self._do_remove_ha_hsrp_asr1k(sub_interface, group, hsrp_ip,
                                      is_primary=is_primary)

    def _do_add_ha_hsrp_asr1k(self, sub_interface, priority, group, vip, vlan,
                              is_primary=True):
        if is_primary is True:
            conf_str = asr1k_snippets.SET_INTC_ASR_HSRP_EXTERNAL % (
                sub_interface, group, priority, group, vip, group, group,
                group, vlan)
            self._edit_running_config(conf_str, 'SET_INTC_ASR_HSRP_EXTERNAL')
        else:
            conf_str = asr1k_snippets.SET_INTC_ASR_SECONDARY_HSRP_EXTERNAL % (
                sub_interface, group, vip)
            self._edit_running_config(conf_str,
                                      'SET_INTC_ASR_SECONDARY_HSRP_EXTERNAL')

    def _do_remove_ha_hsrp_asr1k(self, sub_interface, group, ip,
                                 is_primary=True):
        if is_primary is False:
            conf_str = (asr1k_snippets.REMOVE_INTC_ASR_SECONDARY_HSRP_EXTERNAL
                        % (sub_interface, group, ip))
            self._edit_running_config(
                conf_str, 'REMOVE_INTC_ASR_SECONDARY_HSRP_EXTERNAL')
        else:
            LOG.debug('Standby IP address is not secondary so not removing '
                      'that configuration')

    def _create_sub_interface_v6(self, ri, port, is_external=False, gw_ip=""):
        if self._v6_port_needs_config(port) is not True:
            return
        vrf_name = self._get_vrf_name(ri)
        ip_cidr = port['ip_info']['ip_cidr']
        vlan = self._get_interface_vlan_from_hosting_port(port)
        sub_interface = self._get_interface_name_from_hosting_port(port)
        self._do_create_sub_interface_v6(sub_interface, vlan, vrf_name,
                                         ip_cidr, is_external)
        # Always do HSRP
        self._add_ha_HSRP_v6(ri, port, ip_cidr, is_external)

    def _do_create_sub_interface_v6(self, sub_interface, vlan_id, vrf_name,
                                    ip_cidr, is_external=False):
        if is_external is True:
            conf_str = asr1k_snippets.CREATE_SUBINTERFACE_V6_NO_VRF_WITH_ID % (
                sub_interface, self._deployment_id, vlan_id,
                ip_cidr)
        else:
            conf_str = asr1k_snippets.CREATE_SUBINTERFACE_V6_WITH_ID % (
                sub_interface, self._deployment_id, vlan_id,
                vrf_name, ip_cidr)
        self._edit_running_config(conf_str, '%s CREATE_SUBINTERFACE_V6' %
                                  self.hosting_device['name'])

    def _add_default_route_v6(self, ri, gw_ip, gw_port):
        vrf_name = self._get_vrf_name(ri)
        conf_str = asr1k_snippets.SET_DEFAULT_ROUTE_V6_WITH_INTF % (
            vrf_name, gw_ip)
        self._edit_running_config(conf_str, 'SET_DEFAULT_ROUTE_V6_WITH_INTF')

    def _remove_default_route_v6(self, ri, gw_ip, gw_port):
        vrf_name = self._get_vrf_name(ri)
        sub_interface = self._get_interface_name_from_hosting_port(gw_port)
        self._remove_default_static_route_v6(gw_ip, vrf_name, sub_interface)

    def _remove_default_static_route_v6(self, gw_ip, vrf, out_intf):
        conf_str = asr1k_snippets.REMOVE_DEFAULT_ROUTE_V6_WITH_INTF % (
            vrf, gw_ip)
        self._edit_running_config(conf_str,
                                  'REMOVE_DEFAULT_ROUTE_V6_WITH_INTF')

    def _add_ha_HSRP_v6(self, ri, port, ip, is_external=False):
        if self._v6_port_needs_config(port) is not True:
            return
        vlan = self._get_interface_vlan_from_hosting_port(port)
        group = vlan
        asr_ent = self.hosting_device
        priority = asr_ent['order']
        sub_interface = self._get_interface_name_from_hosting_port(port)
        self._set_ha_HSRP_v6(sub_interface, priority, group, is_external)

    def _port_needs_config(self, port):
        #ToDo(Hareesh): Need to check this!
        if not self._port_is_hsrp(port):
            LOG.debug("Ignoring non-HSRP interface")
            return False
        asr_ent = self._get_asr_ent_from_port(port)
        if asr_ent['name'] != self.hosting_device['name']:
            LOG.debug("Ignoring interface for non-target ASR1k")
            return False
        return True

    @staticmethod
    def _port_is_hsrp(port):
        hsrp_types = [bc.constants.DEVICE_OWNER_ROUTER_HA_INTF]
        return port['device_owner'] in hsrp_types

    @staticmethod
    def _is_global_router(ri):
        return (ri.router.get(ROUTER_ROLE_ATTR) ==
                cisco_constants.ROUTER_ROLE_GLOBAL)

    @classmethod
    def _is_port_v6(cls, port):
        cidr = cls._get_item(port['subnets'],
                             port['ip_info']['subnet_id'], 'id')['cidr']
        return netaddr.IPNetwork(cidr).version == 6

    @staticmethod
    def _get_hsrp_grp_num_from_ri(ri):
        return ri.router['ha_info']['group']

    def _add_internal_nw_nat_rules(self, ri, port, ext_port):
        vrf_name = self._get_vrf_name(ri)
        acl_no = self._generate_acl_num_from_port(port)
        internal_cidr = port['ip_info']['ip_cidr']
        internal_net = netaddr.IPNetwork(internal_cidr).network
        net_mask = netaddr.IPNetwork(internal_cidr).hostmask
        inner_itfc = self._get_interface_name_from_hosting_port(port)
        outer_itfc = self._get_interface_name_from_hosting_port(ext_port)
        self._nat_rules_for_internet_access(acl_no, internal_net,
                                            net_mask, inner_itfc,
                                            outer_itfc, vrf_name)

    def _nat_rules_for_internet_access(self, acl_no, network, netmask,
                                       inner_itfc, outer_itfc, vrf_name):
        """Configure the NAT rules for an internal network.

        Configuring NAT rules in the ASR1k is a three step process. First
        create an ACL for the IP range of the internal network. Then enable
        dynamic source NATing on the external interface of the ASR1k for this
        ACL and VRF of the neutron router. Finally enable NAT on the interfaces
        of the ASR1k where the internal and external networks are connected.

        :param acl_no: ACL number of the internal network.
        :param network: internal network
        :param netmask: netmask of the internal network.
        :param inner_itfc: (name of) interface connected to the internal
        network
        :param outer_itfc: (name of) interface connected to the external
        network
        :param vrf_name: VRF corresponding to this virtual router
        :return: True if configuration succeeded
        :raises: networking_cisco.plugins.cisco.cfg_agent.cfg_exceptions.
        IOSXEConfigException
        """
        acl_present = self._check_acl(acl_no, network, netmask)
        if not acl_present:
            conf_str = snippets.CREATE_ACL % (acl_no, network, netmask)
            self._edit_running_config(conf_str, 'CREATE_ACL')

        pool_name = "%s_nat_pool" % vrf_name
        conf_str = asr1k_snippets.SET_DYN_SRC_TRL_POOL % (acl_no, pool_name,
                                                          vrf_name)
        try:
            self._edit_running_config(conf_str, 'SET_DYN_SRC_TRL_POOL')
        except Exception as dyn_nat_e:
            LOG.info("Ignore exception for SET_DYN_SRC_TRL_POOL: %s. "
                     "The config seems to be applied properly but netconf "
                     "seems to report an error.", dyn_nat_e)

        conf_str = snippets.SET_NAT % (inner_itfc, 'inside')
        self._edit_running_config(conf_str, 'SET_NAT')

        conf_str = snippets.SET_NAT % (outer_itfc, 'outside')
        self._edit_running_config(conf_str, 'SET_NAT')

    def _remove_internal_nw_nat_rules(self, ri, ports, ext_port,
                                      intf_deleted=False):
        """
        Removes the NAT rules already configured when an internal network is
        removed.

        :param ri          -- router-info object
        :param ports       -- list of affected ports where network nat rules
                              was affected
        :param ext_port    -- external facing port
        :param intf_deleted-- If True, indicates that the subinterface was
                              deleted.
        """
        acls = []
        # first disable nat in all inner ports
        for port in ports:
            in_itfc_name = self._get_interface_name_from_hosting_port(port)
            acls.append(self._generate_acl_num_from_port(port))
            is_alone = len(port['change_details']['current_ports']) == 1
            if not intf_deleted and is_alone is True:
                self._remove_interface_nat(in_itfc_name, 'inside')
        # There is a possibility that the dynamic NAT rule cannot be removed
        # from the running config, if there is still traffic in the inner
        # interface causing a rule to be present in the NAT translation
        # table. For this we give 2 seconds for the 'inside NAT rule' to
        # expire and then clear the NAT translation table manually. This can
        # be costly and hence is not enabled here, pending further
        # sinvestigation.

        # LOG.debug("Sleep for 2 seconds before clearing NAT rules")
        # time.sleep(2)
        # clear the NAT translation table
        # self._remove_dyn_nat_translations()

        # remove dynamic nat rules and acls
        vrf_name = self._get_vrf_name(ri)
        ext_itfc_name = self._get_interface_name_from_hosting_port(ext_port)
        for acl in acls:
            self._remove_dyn_nat_rule(acl, ext_itfc_name, vrf_name)

    def _remove_dyn_nat_rule(self, acl_no, outer_itfc_name, vrf_name):
        try:
            pool_name = "%s_nat_pool" % (vrf_name)
            confstr = (asr1k_snippets.REMOVE_DYN_SRC_TRL_POOL %
                (acl_no, pool_name, vrf_name))
            self._edit_running_config(confstr, 'REMOVE_DYN_SRC_TRL_POOL')
        except cfg_exc.IOSXEConfigException as cse:
            LOG.error("temporary disable REMOVE_DYN_SRC_TRL_POOL"
                      " exception handling: %s", (cse))

        conf_str = snippets.REMOVE_ACL % acl_no
        self._edit_running_config(conf_str, 'REMOVE_ACL')

    def _add_floating_ip_asr1k(self, ri, ex_gw_port, floating_ip, fixed_ip):
        vrf_name = self._get_vrf_name(ri)
        self._do_add_floating_ip_asr1k(floating_ip, fixed_ip, vrf_name,
                                       ex_gw_port)

    def _do_add_floating_ip_asr1k(self, floating_ip, fixed_ip, vrf,
                                  ex_gw_port):
        """
        To implement a floating ip, an ip static nat is configured in the
        underlying router ex_gw_port contains data to derive the vlan
        associated with related subnet for the fixed ip.  The vlan in turn
        is applied to the redundancy parameter for setting the IP NAT.
        """
        vlan = ex_gw_port['hosting_info']['segmentation_id']
        hsrp_grp = ex_gw_port[ha.HA_INFO]['group']

        LOG.debug("add floating_ip: %(fip)s, fixed_ip: %(fixed_ip)s, "
                  "vrf: %(vrf)s, ex_gw_port: %(port)s",
                  {'fip': floating_ip, 'fixed_ip': fixed_ip, 'vrf': vrf,
                   'port': ex_gw_port})

        confstr = (asr1k_snippets.SET_STATIC_SRC_TRL_NO_VRF_MATCH %
            (fixed_ip, floating_ip, vrf, hsrp_grp, vlan))
        self._edit_running_config(confstr, 'SET_STATIC_SRC_TRL_NO_VRF_MATCH')

    def _remove_floating_ip(self, ri, ext_gw_port, floating_ip, fixed_ip):
        vrf_name = self._get_vrf_name(ri)
        self._asr_do_remove_floating_ip(floating_ip,
                                        fixed_ip,
                                        vrf_name,
                                        ext_gw_port)

    def _asr_do_remove_floating_ip(self, floating_ip,
                                   fixed_ip, vrf, ex_gw_port):
        vlan = ex_gw_port['hosting_info']['segmentation_id']
        hsrp_grp = ex_gw_port[ha.HA_INFO]['group']

        confstr = (asr1k_snippets.REMOVE_STATIC_SRC_TRL_NO_VRF_MATCH %
            (fixed_ip, floating_ip, vrf, hsrp_grp, vlan))
        self._edit_running_config(confstr,
                                  'REMOVE_STATIC_SRC_TRL_NO_VRF_MATCH')



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_messaging
import six

from networking_cisco.plugins.cisco.common import cisco_constants as const


class DeviceMgrCfgRpcCallback(object):
    """Cisco cfg agent rpc support in Device mgr service plugin."""

    target = oslo_messaging.Target(version='1.1')

    def __init__(self, plugin):
        self._dmplugin = plugin

    def report_non_responding_hosting_devices(self, context, host,
                                              hosting_device_ids):
        """Report that a hosting device is determined to be dead.

        :param context: contains user information
        :param host: originator of callback
        :param hosting_device_ids: list of non-responding hosting devices
        """
        # let the generic status update callback function handle this callback
        self.update_hosting_device_status(context, host,
                                          {const.HD_DEAD: hosting_device_ids})

    def register_for_duty(self, context, host):
        """Report that Cisco cfg agent is ready for duty.

        This function is supposed to be called when the agent has started,
        is ready to take on assignments and before any callbacks to fetch
        logical resources are issued.

        :param context: contains user information
        :param host: originator of callback
        :returns: True if successfully registered, False if not successfully
            registered, None if no handler found. If unsuccessful the agent
            should retry registration a few seconds later
        """
        # schedule any non-handled hosting devices
        return self._dmplugin.auto_schedule_hosting_devices(context, host)

    # version 1.1
    def update_hosting_device_status(self, context, host, status_info):
        """Report status changes for hosting devices.

        :param context: contains user information
        :param host: originator of callback
        :param status_info: Dictionary with list of hosting device ids for
            each type of hosting device status to be updated i.e.::

                {
                    HD_ACTIVE: list_of_ids_of_active_hds,
                    HD_NOT_RESPONDING: list_of_ids_of_not_responding_hds,
                    HD_DEAD: list_of_ids_of_dead_hds,
                    ...
                }
        """
        for status, hd_ids in six.iteritems(status_info):
            # update hosting device entry in db to new status
            hd_spec = {'hosting_device': {'status': status}}
            for hd_id in hd_ids:
                self._dmplugin.update_hosting_device(context, hd_id, hd_spec)
            if status == const.HD_DEAD or status == const.HD_ERROR:
                self._dmplugin.handle_non_responding_hosting_devices(
                    context, host, hd_ids)
            # For status == const.HD_NOT_RESPONDING we do nothing more here as
            # we'll later act on the dead or faulty status change.
            # For status == const.HD_ACTIVE the change in status to active will
            # make the router scheduler consider the now active hosting devices
            # and that is enough so we do nothing more here.

    def get_hosting_devices_for_agent(self, context, host):
        """Fetches routers that a Cisco cfg agent is managing.

        This function is supposed to be called when the agent has started,
        is ready to take on assignments and before any callbacks to fetch
        logical resources are issued.

        :param context: contains user information
        :param host: originator of callback
        :returns: dict of hosting devices managed by the cfg agent
        """
        agent_ids = self._dmplugin.get_cfg_agents(context, active=None,
                                                  filters={'host': [host]})
        if agent_ids:
            return [self._dmplugin.get_device_info_for_agent(context, hd_db)
                    for hd_db in self._dmplugin.get_hosting_devices_db(
                    context, filters={'cfg_agent_id': [agent_ids[0].id]})]
        return []
